import nctoolkit as nc
import re
import pandas as pd
import os
import glob
import warnings
from oceanval.session import session_info

session_info["keys"] = []

def find_recipe(x, start = None, end = None):
    output = dict()
    # check if there is only one key and one value
    if len(x.keys()) != 1:
        raise ValueError("Recipe dictionary must have exactly one key") 
    if len(x.values()) != 1:
        raise ValueError("Recipe dictionary must have exactly one value") 
    
    valid_recipes = dict()
    valid_recipes["nitrate"] = "nsbc"

    output["vertical"] = None

    name = x.keys()
    # first key
    name = list(name)[0]
    value = x[name]
    # add a suitable short name
    if name == "chlorophyll":
        output["short_name"] = "chlorophyll concentration"
    if name == "oxygen":
        output["short_name"] = "dissolved oxygen"
    if name == "temperature":
        output["short_name"] = "sea temperature"
    if name == "salinity":
        output["short_name"] = "salinity"
    if name == "nitrate":
        output["short_name"] = "nitrate concentration"
    if name == "ammonium":
        output["short_name"] = "ammonium concentration"
    if name == "phosphate":
        output["short_name"] = "phosphate concentration"
    if name == "silicate":
        output["short_name"] = "silicate concentration"
    # add a long name
    if name == "chlorophyll":
        output["long_name"] = "chlorophyll a concentration"
    if name == "oxygen":
        output["long_name"] = "dissolved oxygen concentration"
    if name == "temperature":
        output["long_name"] = "sea water temperature"
    if name == "salinity":
        output["long_name"] = "sea water salinity"
    if name == "nitrate":
        output["long_name"] = "nitrate concentration"
    if name == "ammonium":
        output["long_name"] = "ammonium concentration"
    if name == "phosphate":
        output["long_name"] = "phosphate concentration"
    if name == "silicate":
        output["long_name"] = "silicate concentration"
    # add title
    if name == "chlorophyll":
        output["short_title"] = "Chlorophyll"
    if name == "oxygen":
        output["short_title"] = "Oxygen"
    if name == "temperature":
        output["short_title"] = "Temperature"
    if name == "salinity":
        output["short_title"] = "Salinity"
    if name == "nitrate":
        output["short_title"] = "Nitrate"
    if name == "ammonium":
        output["short_title"] = "Ammonium"
    if name == "phosphate":
        output["short_title"] = "Phosphate"
    if name == "silicate":
        output["short_title"] = "Silicate"

    if name == "kd490":
        output["short_name"] = "KD490"
        output["short_title"] = "KD490"
        output["long_name"] = "diffuse attenuation coefficient at 490 nm"
    # COBE2 temperature options

    if name.lower() == "ph":
        output["short_name"] = "pH"
        output["long_name"] = "sea water pH"
        output["short_title"] = "pH"
    if name.lower() == "alkalinity":
        output["short_name"] = "total alkalinity"
        output["long_name"] = "sea water total alkalinity"
        output["short_title"] = "Total Alkalinity"
    
    if value.lower() == "glodap":
        if name.lower() == "ph":
            output["obs_path"]=  "https://www.ncei.noaa.gov/data/oceans/archive/arc0221/0286118/1.1/data/0-data/GLODAPv2.2016b_MappedClimatologies/GLODAPv2.2016b.pHtsinsitutp.nc"
            output["source"] = "GLODAPv2.2016b"
            output["source_info"] = "Lauvset, S. K., Key, R. M., Olsen, A., van Heuven, S., Velo, A., Lin, X., Schirnick, C., Kozyr, A., Tanhua, T., Hoppema, M., Jutterström, S., Steinfeldt, R., Jeansson, E., Ishii, M., Perez, F. F., Suzuki, T., and Watelet, S.: A new global interior ocean mapped climatology: the 1° ×  1° GLODAP version 2, Earth Syst. Sci. Data, 8, 325–340, https://doi.org/10.5194/essd-8-325-2016, 2016."
            output["name"] = name
            output["thredds"] = False
            output["climatology"] = True
            output["thredds"] = True
            output["obs_variable"] = 'pHtsinsitutp'
            return output
        if name.lower() == "alkalinity":
            output["obs_path"]=  "https://www.ncei.noaa.gov/data/oceans/archive/arc0221/0286118/1.1/data/0-data/GLODAPv2.2016b_MappedClimatologies/GLODAPv2.2016b.TAlk.nc"
            output["source"] = "GLODAPv2.2016b"
            output["source_info"] = "Lauvset, S. K., Key, R. M., Olsen, A., van Heuven, S., Velo, A., Lin, X., Schirnick, C., Kozyr, A., Tanhua, T., Hoppema, M., Jutterström, S., Steinfeldt, R., Jeansson, E., Ishii, M., Perez, F. F., Suzuki, T., and Watelet, S.: A new global interior ocean mapped climatology: the 1° ×  1° GLODAP version 2, Earth Syst. Sci. Data, 8, 325–340, https://doi.org/10.5194/essd-8-325-2016, 2016."
            output["name"] = name
            output["thredds"] = False
            output["climatology"] = True
            output["thredds"] = True
            output["obs_variable"] = 'TAlk'
            return output



    
    if value.lower() == "cobe2":
        if name.lower() == "temperature": 
            url = f"https://psl.noaa.gov/thredds/dodsC/Datasets/COBE2/sst.mon.mean.nc"
            output["obs_path"] = url
            output["source"] = "COBE2"
            output["source_info"] = "COBE-SST 2 and Sea Ice data provided by the NOAA PSL, Boulder, Colorado, USA, from their website at https://psl.noaa.gov/data/gridded/data.cobe2.html."
            output["name"] = name
            output["thredds"] = True
            output["climatology"] = False
            output["vertical"] = False

            return output 

    if value.lower() == "woa23":
        output["source"] = "WOA23"
        output["source_info"] = " Garcia, H.E., C. Bouchard, S.L. Cross, C.R. Paver, Z. Wang, J.R. Reagan, T.P. Boyer, R.A. Locarnini, A.V. Mishonov, O. Baranova, D. Seidov, and D. Dukhovskoy. World Ocean Atlas 2023, Volume 4: Dissolved Inorganic Nutrients (phosphate, nitrate, silicate). A. Mishonov, Tech. Ed. NOAA Atlas NESDIS 92, doi.org/10.25923/39qw-7j08"
        output["climatology"] = True
        output["name"] = name
        output["thredds"] = True

        if name.lower() == "nitrate":
            url = []
            for month in range(1,13):
                # format month to two digits
                month_str = f"{month:02d}"
                url.append(f"https://www.ncei.noaa.gov/thredds-ocean/dodsC/woa23/DATA/nitrate/netcdf/all/1.00/woa23_all_n{month_str}_01.nc")
            output["obs_path"] = url
            output["obs_variable"] = "n_an"
            return output

        # now do oxygen

        if name.lower() == "oxygen":
            url = []
            for month in range(1,13):
                # format month to two digits
                month_str = f"{month:02d}"
                #https://www.ncei.noaa.gov/thredds-ocean/dodsC/woa23/DATA/oxygen/netcdf/all/1.00/woa23_all_o01_01.nc.html
                url.append(f"https://www.ncei.noaa.gov/thredds-ocean/dodsC/woa23/DATA/oxygen/netcdf/all/1.00/woa23_all_o{month_str}_01.nc")
            output["obs_path"] = url
            output["name"] = name
            output["obs_variable"] = "o_an"
            return output
        # phosphate
        if name.lower() == "phosphate":
            output["obs_variable"] = "p_an"
            url = []
            for month in range(1,13):
                # format month to two digits
                month_str = f"{month:02d}"
                url.append(f"https://www.ncei.noaa.gov/thredds-ocean/dodsC/woa23/DATA/phosphate/netcdf/all/1.00/woa23_all_p{month_str}_01.nc")
            output["obs_path"] = url
            return output
        
        # silicate
        if name.lower() == "silicate":
            output["obs_variable"] = "i_an"
            url = []
            for month in range(1,13):
                # format month to two digits
                month_str = f"{month:02d}"
                url.append(f"https://www.ncei.noaa.gov/thredds-ocean/dodsC/woa23/DATA/silicate/netcdf/all/1.00/woa23_all_i{month_str}_01.nc")
            output["obs_path"] = url
            return output
        # todo
        # salinity
        # temperature
        if name.lower() in ["salinity", "temperature"]:
            # check if start and end are provided
            if start is None or end is None:

                valid_periods = [""]
                raise ValueError("Start and end depth must be provided for salinity and temperature WOA23 recipes")
            # valid time period are
            # 1955-1964
            # 1965-1974
            # 1975-1984
            # 1985-1994
            # 1995-2004
            # 2005-2014
            # 2015-2022
            # start and end must fall into one of these time periods
            # first check if they are more than 9 years apart
            if end - start > 9:
                raise ValueError("Start and end depth must fall within a single WOA23 climatological period (10 year periods)")
            # identify the period it is in based on start year
            if start >= 1955 and end <= 1964:
                period = "5564"
            elif start >= 1965 and end <= 1974:
                period = "6574"
            elif start >= 1975 and end <= 1984:
                period = "7584"
            elif start >= 1985 and end <= 1994:
                period = "8594"
            elif start >= 1995 and end <= 2004:
                period = "95A4"
            elif start >= 2005 and end <= 2014:
                period = "A5B4"
            elif start >= 2015 and end <= 2022:
                period = "B5C2"
            if end > 2022:
                raise ValueError("End year cannot be greater than 2022 for WOA23 recipes")

                #https://www.ncei.noaa.gov/thredds-ocean/dodsC/woa23/DATA/temperature/netcdf/5564/1.00/woa23_5564_t00_01.nc.html
            #url = f"https://www.ncei.noaa.gov/thredds-ocean/dodsC/woa23/DATA/{name.lower()}/netcdf/{period.replace('-', '')}/1.00/woa23_{period.replace('-', '')}_{name[0].lower()}00_01.nc"
            urls = []
            for month in range(1,13):
                month_str = f"{month:02d}"
                #https://www.ncei.noaa.gov/thredds-ocean/dodsC/woa23/DATA/temperature/netcdf/5564/1.00/woa23_5564_t00_01.nc.html
                urls.append(f"https://www.ncei.noaa.gov/thredds-ocean/dodsC/woa23/DATA/{name.lower()}/netcdf/{period}/1.00/woa23_{period}_{name[0].lower()}{month_str}_01.nc")
                #urls.append(url)

            output["obs_path"] = urls
            output["name"] = name
            if name.lower() == "salinity":
                output["obs_variable"] = "s_an"
            if name.lower() == "temperature":
                output["obs_variable"] = "t_an"
            return output

    if value == "occci":
        urls = []
        if name == "chlorophyll":
            for yy in range(1998, 2025):
                for month in range(1, 13):
                    month_str = f"{month:02d}"
                    url = f"https://www.oceancolour.org/thredds/dodsC/cci/v6.0-release/geographic/monthly/chlor_a/{yy}/ESACCI-OC-L3S-CHLOR_A-MERGED-1M_MONTHLY_4km_GEO_PML_OCx-{yy}{month_str}-fv6.0.nc"
                    urls.append(url)
            output["obs_path"] = urls
            output["source"] = "OCCCI"
            output["source_info"] = "Sathyendranath, S, Brewin, RJW, Brockmann, C, Brotas, V, Calton, B, Chuprin, A, Cipollini, P, Couto, AB, Dingle, J, Doerffer, R, Donlon, C, Dowell, M, Farman, A, Grant, M, Groom, S, Horseman, A, Jackson, T, Krasemann, H, Lavender, S, Martinez-Vicente, V, Mazeran, C, Mélin, F, Moore, TS, Müller, D, Regner, P, Roy, S, Steele, CJ, Steinmetz, F, Swinton, J, Taberner, M, Thompson, A, Valente, A, Zühlke, M, Brando, VE, Feng, H, Feldman, G, Franz, BA, Frouin, R, Gould, Jr., RW, Hooker, SB, Kahru, M, Kratzer, S, Mitchell, BG, Muller-Karger, F, Sosik, HM, Voss, KJ, Werdell, J, and Platt, T (2019) An ocean-colour time series for use in climate studies: the experience of the Ocean-Colour Climate Change Initiative (OC-CCI). Sensors: 19, 4285. doi:10.3390/s19194285."
            # short name
            output["short_name"] = "chlorophyll concentration"
            output["name"] = name
            output["obs_variable"] = "chlor_a"
            output["thredds"] = True
            output["climatology"] = False
            return output

    # kd490
    if name == "kd490":
        for yy in range(1998, 2025):
            for month in range(1, 13):
                month_str = f"{month:02d}"
                #https://www.oceancolour.org/thredds/dodsC/cci/v6.0-release/geographic/monthly/kd/1997/ESACCI-OC-L3S-K_490-MERGED-1M_MONTHLY_4km_GEO_PML_KD490_Lee-199709-fv6.0.nc.html
                url = f"https://www.oceancolour.org/thredds/dodsC/cci/v6.0-release/geographic/monthly/kd/{yy}/ESACCI-OC-L3S-K_490-MERGED-1M_MONTHLY_4km_GEO_PML_KD490_Lee-{yy}{month_str}-fv6.0.nc"
                urls.append(url)
        output["obs_path"] = urls
        output["source"] = "OCCCI"
        output["source_info"] = "Sathyendranath, S, Brewin, RJW, Brockmann, C, Brotas, V, Calton, B, Chuprin, A, Cipollini, P, Couto, AB, Dingle, J, Doerffer, R, Donlon, C, Dowell, M, Farman, A, Grant, M, Groom, S, Horseman, A, Jackson, T, Krasemann, H, Lavender, S, Martinez-Vicente, V, Mazeran, C, Mélin, F, Moore, TS, Müller, D, Regner, P, Roy, S, Steele, CJ, Steinmetz, F, Swinton, J, Taberner, M, Thompson, A, Valente, A, Zühlke, M, Brando, VE, Feng, H, Feldman, G, Franz, BA, Frouin, R, Gould, Jr., RW, Hooker, SB, Kahru, M, Kratzer, S, Mitchell, BG, Muller-Karger, F, Sosik, HM, Voss, KJ, Werdell, J, and Platt, T (2019) An ocean-colour time series for use in climate studies: the experience of the Ocean-Colour Climate Change Initiative (OC-CCI). Sensors: 19, 4285. doi:10.3390/s19194285."
        output["name"] = name
        output["obs_variable"] = "kd_490"
        output["thredds"] = True
        output["climatology"] = False
        return output


    
    if value.lower() == "nsbc":
        if name.lower() in ["ammonium", "nitrate", "phosphate", "silicate", "chlorophyll", "oxygen", "temperature", "salinity"]:
            if name.lower() == "chlorophyll":
                url = f"https://icdc.cen.uni-hamburg.de/thredds/dodsC/ftpthredds/nsbc/level_3/climatological_monthly_mean/NSBC_Level3_{'chlorophyll_a'}__UHAM_ICDC__v1.1__0.25x0.25deg__OAN_1960_2014.nc"
            else:
                url = f"https://icdc.cen.uni-hamburg.de/thredds/dodsC/ftpthredds/nsbc/level_3/climatological_monthly_mean/NSBC_Level3_{name}__UHAM_ICDC__v1.1__0.25x0.25deg__OAN_1960_2014.nc"
            output["obs_path"] = url
            output["source"] = "NSBC"
            output["source_info"] = "Hinrichs, Iris; Gouretski, Viktor; Paetsch, Johannes; Emeis, Kay; Stammer, Detlef (2017). North Sea Biogeochemical Climatology (Version 1.1)."
            output["name"] = name
            output["thredds"] = True
            output["climatology"] = True
            if name.lower() == "ammonium":
                output["obs_variable"] = "ammonium_mean"
            if name.lower() == "nitrate":
                output["obs_variable"] = "nitrate_mean"
            if name.lower() == "phosphate":
                output["obs_variable"] = "phosphate_mean"
            if name.lower() == "silicate":
                output["obs_variable"] = "silicate_mean"
            if name.lower() == "chlorophyll":
                output["obs_variable"] = "chlorophyll_a_mean"
            if name.lower() == "oxygen":
                output["obs_variable"] = "oxygen_mean"
            if name.lower() == "temperature":
                output["obs_variable"] = "temperature_mean"
            if name.lower() == "salinity":
                output["obs_variable"] = "salinity_mean"



            return output

    raise ValueError(f"Recipe value {value} is not valid for recipe name {name}")


    return x

def get_name(obj):
    namespace = globals()
    for name in namespace:
        if namespace[name] is obj:
            return name
    return None

# create a validator class
# create a variable class to hold metadata
class Variable:
    def __init__(self):
        self.long_name = None
        self.gridded = None 
    def __str__(self):
    # add a print method for each atrribute
        attrs = vars(self)
        return '\n'.join("%s: %s" % item for item in attrs.items())  
    # add a repr method
    def __repr__(self):
        attrs = vars(self)
        return '\n'.join("%s: %s" % item for item in attrs.items()) 




class Validator:
    def __init__(self):
        self.rules = {}

    def validate(self, data):
        errors = {}
        for field, rule in self.rules.items():
            if field not in data:
                errors[field] = "Field is missing"
            elif not rule(data[field]):
                errors[field] = "Validation failed"
        return errors
    # add chlorophyll
    def add_rule(self, field, rule):
        self.rules[field] = rule

    keys = session_info["keys"]
    # add a deleter that removes from keys list
    def __delattr__(self, name):
        if name != "keys" and "rules" not in name:
            if name in self.keys:
                self.keys.remove(name)
        super().__delattr__(name)
    # add remove method
    def remove(self, name):
        if name != "keys" and "rules" not in name:
            if name in self.keys:
                self.keys.remove(name)
        super().__delattr__(name)


    # add reset method
    def reset(self):
        for key in self.keys:
            # do not remove rules
            if "rules" not in key:
                super().__delattr__(key)
        self.keys = []

    # ensure self.x = y, adds x to the keys list
    def __setattr__(self, name, value):
        if name != "keys" and "rules" not in name:
            if name not in self.keys:
                self.keys.append(name)
                # ensure this can be accessed via self[name]

        super().__setattr__(name, value)

    # create a [] style accessor
    # make Validator subsettable, so that validator["chlorophyll"] returns the chlorophyll variable
    def __getitem__(self, key):
        # if key == "chlorophyll":
        #     return  
        return getattr(self, key, None)
    

    
    # 
    def add_gridded_comparison(self, 
                               name = None, 
                               long_name = None, 
                               short_name = None, 
                               short_title = None, 
                               source = None, 
                               source_info = None, 
                               model_variable = None, 
                               obs_path = None, 
                               obs_variable = "auto", 
                               start = -1000, 
                               end = 3000, 
                               vertical = False, 
                               climatology = None, 
                               obs_multiplier = 1,
                               obs_adder = 0,
                               thredds = False,
                               recipe = None
                                   ): 
        """

        Add a gridded comparison variable to the Validator

        Parameters:

        name (str): Name of the variable

        long_name (str): Long name of the variable

        short_name (str): Short name of the variable

        short_title (str): Short title of the variable

        source (str): Source of the variable

        source_info (str): Source information of the variable

        model_variable (str): Model variable name

        obs_path (str): Directory or path of the observations

        obs_variable (str): Observation variable name

        start (int): Start depth of the variable

        end (int): End depth of the variable

        vertical (bool): Whether the variable is vertical

        climatology (bool): Whether to use climatology

        obs_multiplier (float): Multiplier for the observation

        obs_adder (float): Adder for the observation

        """

        # maybe include an averaging option: daily, monthly, annual etc.


        if recipe is not None:
            recipe_info = find_recipe(recipe, start = start, end = end)
            obs_path = recipe_info["obs_path"]
            source = recipe_info["source"]
            source_info = recipe_info["source_info"]
            thredds = recipe_info["thredds"]
            climatology = recipe_info["climatology"]
            name = recipe_info["name"]
            short_name = recipe_info["short_name"]
            long_name = recipe_info["long_name"]
            short_title = recipe_info["short_title"]
            # vertical is not None
            if recipe_info["vertical"] is not None:
                vertical = recipe_info["vertical"]
            try:
                obs_variable = recipe_info["obs_variable"]
            except:
                pass
            recipe = True
        else:
            recipe = False
        
        # name can only have str or numbers
        if not re.match("^[A-Za-z0-9]+$", name):
            raise ValueError("Name can only contain letters and numbers")


        if source is None:
            raise ValueError("Source must be supplied")
        if model_variable is None:
            raise ValueError("Model variable must be supplied")
        # climatology must be provideded
        if climatology is None:
            raise ValueError("Climatology must be provided for gridded comparison")
        # obs_path is needed
        if obs_path is None:
            raise ValueError("obs_path must be provided for gridded comparison")
        # must be boolean
        if not isinstance(climatology, bool):
            raise ValueError("Climatology must be a boolean value")
        try:
            obs_multiplier  = float(obs_multiplier)
        except:
            raise ValueError("obs_multiplier must be a number")

        try:
            obs_adder  = float(obs_adder)
        except:
            raise ValueError("obs_adder must be a number")

        assumed = []

        if long_name is None:
            try:
                long_name = self[name].long_name
            except:
                long_name = name
                assumed.append("long_name")

        if short_name is None:
            # use it, if it already exists
            try:
                short_name = self[name].short_name
            except:
                short_name = name
                assumed.append("short_name")
        if short_title is None:
            try:
                short_title = self[name].short_title
            except:
                short_title = name.title()
                assumed.append("short_title")


        if source_info is None:
            source_info = f"Source for {source}"
            assumed.append("source_info")

        source_name = source
        source = {source: source_info}
        # ensure the sourc key does not included "_"
        if "_" in list(source_name)[0]:
            raise ValueError("Source cannot contain '_'")
        if not isinstance(obs_variable, str):
            raise ValueError("obs_variable must be a string")

        gridded_dir = obs_path
        if gridded_dir != "auto":
            if thredds is False:
                if not os.path.exists(gridded_dir):
                    raise ValueError(f"Gridded directory {gridded_dir} does not exist")
        # thredds must be boolean
        if not isinstance(thredds, bool):
            raise ValueError("thredds must be a boolean value")

        if name in session_info["short_title"]:
            if short_title is not None:
                if short_title != session_info["short_title"][name]:
                    raise ValueError(f"Short title for {name} already exists as {session_info['short_title'][name]}, cannot change to {short_title}")





        # Figure out if name is already 

        # figure out if self[name] exists already
        if getattr(self, name, None) is None:
            var = Variable()
            setattr(self, name, var)
            self[name].point_dir = None
            self[name].point_source = None
            self[name].sources = source 
            self[name].point_start = -1000
            self[name].point_end = 3000
            self[name].vertical_point = None
            self[name].model_variable = None
            self[name].obs_multiplier = 1
            self[name].binning = None
            self[name].climatology = None
            self[name].sources = dict()

        else:
            if self[name].model_variable != model_variable:
                raise ValueError(f"Model variable for {name} already exists as {old_model_variable}, cannot change to {model_variable}")
            if self[name].sources is not None:
                orig_sources = self[name].sources
            if list(source.keys())[0] in orig_sources:
                # ensure the value is the same
                if orig_sources[list(source.keys())[0]] != source[list(source.keys())[0]]:
                    raise ValueError(f"Source {list(source.keys())[0]} already exists with a different value")

        self[name].sources[source_name] = source_info
        self[name].obs_adder_gridded = obs_adder
        self[name].thredds = thredds
        self[name].climatology = climatology
        self[name].obs_multiplier_gridded = obs_multiplier
        self[name].n_levels = 1
        self[name].vertical_gridded = vertical
        self[name].gridded_start = start
        self[name].gridded_end = end
        self[name].gridded = True
        self[name].long_name = long_name
        # if this is None set to Name
        self[name].short_name = short_name
        if self[name].short_name is None:
            self[name].short_name = name
            assumed.append("short_name")    
        self[name].short_title = short_title
        if self[name].short_title is None:
            self[name].short_title = name.title()
            assumed.append("short_title")
        # check if this is c
        session_info["short_title"][name] = self[name].short_title

        self[name].sources[source_name] = source_info 
        self[name].gridded_source = list(source.keys())[0]
        self[name].model_variable = model_variable
        # add obs_variable, ensure it's a string
        self[name].obs_variable = obs_variable
        # check this exists
        gridded_dir = obs_path
        self[name].gridded_dir = gridded_dir
        self[name].recipe = recipe
        
        # ensure nothing is None
        # warnings for assumptions
        if len(assumed) > 0:
            print(f"Warning: The following attributes were missing and were assumed for variable {name}: {assumed}")


    def add_point_comparison(self, 
                             name = None, 
                             long_name = None, 
                             vertical = False, 
                             short_name = None, 
                             short_title = None, 
                             source = None, 
                             source_info = None, 
                             model_variable = None, 
                             start = -1000, 
                             end = 3000, 
                             obs_path = None, 
                             obs_multiplier = 1, 
                             obs_adder = 0,
                             binning = None  ):
        """

        Add a point comparison variable to the Validator

        Parameters:

        name (str): Name of the variable

        long_name (str): Long name of the variable

        vertical (bool): Whether the variable is vertical

        short_name (str): Short name of the variable

        short_title (str): Short title of the variable

        source (str): Source of the variable

        source_info (str): Source information of the variable

        model_variable (str): Model variable name

        start (int): Start depth of the variable

        end (int): End depth of the variable

        obs_path (str): Directory of the observations

        obs_multiplier (float): Multiplier for the observation, if needed to convert units

        binning (list): Binning information [lon_resolution, lat_resolution]

        """

        # check what is supplied is valid
        # name can only have str or numbers
        if not re.match("^[A-Za-z0-9]+$", name):
            raise ValueError("Name can only contain letters and numbers")

        if source is None:
            raise ValueError("Source must be supplied")
        source_name = source
        # ensure the sourc key does not included "_"
        if "_" in source: 
            raise ValueError("Source key cannot contain '_'")


        try:
            obs_multiplier= float(obs_multiplier)
        except:
            raise ValueError("obs_multiplier must be a number")
        try:
            obs_adder = float(obs_adder)
        except:
            raise ValueError("obs_adder must be a number")
        # vertical must be a boolean
        if not isinstance(vertical, bool):
            raise ValueError("vertical must be a boolean value")

        # check these are int or can be cast to int
        try:
            start = int(start)
            end = int(end)
        except:
            raise ValueError("start and end must be integers")

        assumed = []
        if source_info is None:
            source_info = f"Source for {source}"
            assumed.append("source_info")
        source = {source_name: source_info}
        if long_name is None:
            try:
                long_name = self[name].long_name
            except:
                long_name = name
                assumed.append("long_name")
        if short_name is None:
            # use it, if it already exists
            try:
                short_name = self[name].short_name
            except:
                short_name = name
                assumed.append("short_name")
        if short_title is None:
            try:
                short_title = self[name].short_title
            except:
                short_title = name.title()
                assumed.append("short_title")

        if name in session_info["short_title"]:
            if short_title != session_info["short_title"][name]:
                raise ValueError(f"Short title for {name} already exists as {session_info['short_title'][name]}, cannot change to {short_title}")

        point_files = [f for f in glob.glob(os.path.join(obs_path, "*.csv"))] 
        # if no files exists, raise error
        if len(point_files) == 0:
            raise ValueError(f"No csv files found in point directory {obs_path}")
        valid_vars = ["lon", "lat", "year", "month", "day", "depth", "observation", "source"]
        vertical_option = False
        for vv in point_files:
            # read in the first row
            df = pd.read_csv(vv, nrows=1)
            # throw error something else is in there
            bad_cols = [col for col in df.columns if col not in valid_vars]
            if len(bad_cols) > 0:
                raise ValueError(f"Invalid columns {bad_cols} found in point data file {vv}")
            if "depth" in df.columns:
                vertical_option = True
            # lon/lat/observation *must* be in df
            for req_col in ["lon", "lat", "observation"]:
                if req_col not in df.columns:
                    raise ValueError(f"Required column {req_col} not found in point data file {vv}")
        if vertical_option is False:
            if vertical:
                raise ValueError("vertical is set to True but no depth column found in point data files. You cannot vertically validate this data.")
        # if binning is supplied, ensure it is a 2 variable list
        if binning is not None:
            if not isinstance(binning, list) or len(binning) != 2:
                raise ValueError("binning must be a list of two values: [spatial_resolution, depth_resolution]")
        # ensure each element of binning is a number
            for res in binning:
                try:
                    float(res)
                except:
                    raise ValueError("Each element of binning must be a number")
        # check this exists
        point_dir = obs_path
        if point_dir != "auto":
            if not os.path.exists(point_dir):
                raise ValueError(f"Point directory {point_dir} does not exist")


        # figure out if self[name] exists already
        if getattr(self, name, None) is None:
            # add it
            var = Variable()
            setattr(self, name, var)
            self[name].gridded = False
            self[name].vertical_gridded = None 
            self[name].recipe = None 
            self[name].sources = dict() 
            self[name].gridded_source = None 
            self[name].thredds = None 
            self[name].gridded_dir = None 
            self[name].obs_variable = None 

        else:

            # ensure short title is the same
            if short_title != session_info["short_title"][name]:
                raise ValueError(f"Short title for {name} already exists as {session_info['short_title'][name]}, cannot change to {short_title}")

            if self[name].model_variable != model_variable:
                raise ValueError(f"Model variable for {name} already exists as {old_model_variable}, cannot change to {model_variable}")
            if self[name].sources is not None:
                orig_sources = self[name].sources
            if list(source.keys())[0] in orig_sources:
                # ensure the value is the same
                if orig_sources[list(source.keys())[0]] != source[list(source.keys())[0]]:
                    raise ValueError(f"Source {list(source.keys())[0]} already exists with a different value")

        self[name].sources[source_name] = source_info

        self[name].obs_multiplier_point= obs_multiplier
        self[name].obs_adder_point = obs_adder
        self[name].n_levels = 1
        self[name].long_name = long_name
        if self[name].long_name is None:
            self[name].long_name = name
            assumed.append("long_name")

        self[name].vertical_point = vertical
        self[name].short_name = short_name
        if self[name].short_name is None:
            self[name].short_name = name
            assumed.append("short_name")

        self[name].short_title = short_title
        if self[name].short_title is None:
            self[name].short_title = name.title()
            assumed.append("short_title")
        self[name].point_start = start
        self[name].point_end = end
        # append source to the var.source
        # check if source key is in orig_source
        self[name].point_source = list(source.keys())[0]   
        self[name].model_variable = model_variable
        self[name].point_dir = obs_path

        # figure out if var.binning exists
        self[name].binning = binning 

        #  

        for vv in assumed:
            print(f"Warning: The attribute {vv} was missing and was assumed for variable {name}")
        session_info["short_title"][name] = short_title


definitions = Validator()



def generate_mapping(ds):
    """
    Generate mapping of model and observational variables
    """

    model_dict = {}
    try:
        candidate_variables = definitions.keys
        ds1 = nc.open_data(ds[0], checks=False)
        ds_contents = ds1.contents

        ds_contents["long_name"] = [str(x) for x in ds_contents["long_name"]]

        ds_contents_top = ds_contents.query("nlevels == 1").reset_index(drop=True)
        n_levels = int(ds_contents.nlevels.max())
        if n_levels > session_info["n_levels"]:
            session_info["n_levels"] = n_levels
        # number of rows in ds_contents
        if len(ds_contents) == 0:
            ds_contents = ds_contents_top
    except:
        return model_dict

    for vv in candidate_variables:
        variables = definitions[vv].model_variable.split("+")
        include = True
        for var in variables:
            if var not in ds_contents.variable.values:
                include = False
        if include:
            model_dict[vv] = definitions[vv].model_variable
            n_levels = ds_contents.query("variable in @variables")["nlevels"].max()
            if n_levels > definitions[vv].n_levels:
                definitions[vv].n_levels = n_levels 
            continue

    return model_dict






# %% tags=["remove-input"]


if variable == "chlorophyll":
    transformation = "log10"
else:
    transformation = None 


ds_annual = ds_model.copy()
ds_annual.tmean()
# if "model" == ds_annual.variables[0]:
#     ds_annual.set_longnames({"model": Variable})
# else:
#     ds_annual.set_longnames({variable: Variable})

ds_annual.run()

#get the min/max lon lat with actual values in ds_annual
# save as [lon_min, lon_max] and [lat_min, lat_max]
coord_ranges = (
    ds_annual
    .to_dataframe()
    .reset_index()
    .dropna()
)
lon_name = [x for x in coord_ranges.columns if "lon" in x][0]
lat_name = [x for x in coord_ranges.columns if "lat" in x][0]
coord_ranges = (
    coord_ranges
    # rename the columns to lon and lat
    .rename(columns = {lon_name: "lon", lat_name: "lat"})
    #
    .loc[:,["lon", "lat"]]
    .agg(["min", "max"])
    .to_dict()
)

lon_min = coord_ranges["lon"]["min"]
lon_max = coord_ranges["lon"]["max"]
lat_min = coord_ranges["lat"]["min"]
lat_max = coord_ranges["lat"]["max"]
if lon_max < 90:
    ds_annual.subset(lon = [lon_min, lon_max], lat = [lat_min, lat_max])
fix_grid = False

# %% tags=["remove-cell"]
# Plot them on the same scale
ds_both = ds_model.copy()
ds_both.rename({ds_both.variables[0]: "model"})
ds_both.append(ds_obs)
ds_both.rename({ds_obs.variables[0]: "observation"})
ds_both.merge(match = "month")
ds_both.run()
ds_both.variables

#get the min/max lon lat with actual values in ds_annual
# save as [lon_min, lon_max] and [lat_min, lat_max]
coord_ranges = (
    ds_both
    .to_dataframe()
    .reset_index()
    .dropna()
)
lon_name = [x for x in coord_ranges.columns if "lon" in x][0]
lat_name = [x for x in coord_ranges.columns if "lat" in x][0]
coord_ranges = (
    coord_ranges
    # rename the columns to lon and lat
    .rename(columns = {lon_name: "lon", lat_name: "lat"})
    #
    .loc[:,["lon", "lat"]]
    .agg(["min", "max"])
    .to_dict()
)

lat_min = coord_ranges["lat"]["min"]
lat_max = coord_ranges["lat"]["max"]
if lon_max < 90:
    ds_both.subset(lon = [lon_min, lon_max], lat = [lat_min, lat_max])

if fix_grid:
    ds_both.to_latlon(lon = [lon_min , lon_max], lat = [lat_min, lat_max], res = [lon_res, lat_res])


# %% tags=["remove-cell"]
import matplotlib.pyplot as plt

plt.subplots_adjust(wspace=20, hspace=20)

fig = plt.figure(figsize=(14, 14))

# Create 1x2 Grid

gs = fig.add_gridspec(nrows=1, ncols=2, wspace = 0.35, hspace = 0)

# get limits..

ds_both.tmean()

z_min = (
    ds_both
    .to_dataframe()
    .reset_index()
    .dropna()
    .loc[:,["model", "observation"]]
    # get the 2nd percentils
    # .groupby("variable")
    .quantile(0.02)
    .min()
)

# get the 98th percentils

z_max = (
    ds_both
    .to_dataframe()
    .reset_index()
    .dropna()
    .loc[:,["model", "observation"]]
    .quantile(0.98)
    .max()
)
# fix the units

ds_both.set_units({"observation": ds_both.contents.query("variable == 'model'").reset_index().unit.values[0]})
# chang longname
# try:
#     ds_both.set_longnames({"observation": ds_both.contents.query("variable == 'model'").reset_index().long_name.values[0]}) 
# except:
#     ds_both.set_longnames({"observation": ds_both.contents.query("variable == 'observation'").reset_index().long_name.values[0]})  
# ditch #Modelled and Observed from long names
the_contents = ds_both.contents
for vv in the_contents.variable:
    long_name = the_contents.query(f"variable == '{vv}'").reset_index().long_name.values[0]
    if "Modelled" in long_name:
        long_name = long_name.replace("Modelled", "").strip().replace("surface", "Surface")
    if "Observed" in long_name:
        long_name = long_name.replace("Observed", "").strip().replace("surface", "Surface")
    ds_both.set_longnames({vv: new_name})

try:
    ds_both.pub_plot(variable  = "model", limits = [z_min, z_max], title = "Model", fig = fig, gs = gs[0,0], trans = transformation)
except:
    lon_res = (lon_max - lon_min) / 100
    lat_res = (lat_max - lat_min) / 100
    ds_both.to_latlon(lon = [lon_min , lon_max], lat = [lat_min, lat_max], res = [lon_res, lat_res])
    md("The model grid was regridded to a regular lon-lat grid for plotting purposes.")
    ds_both.pub_plot(variable  = "model", limits = [z_min, z_max], title = "Model", fig = fig, gs = gs[0,0], trans = transformation)



ds_both.pub_plot(variable  = "observation", limits = [z_min, z_max], title = "Observation", fig = fig, gs = gs[0,1], trans = transformation)

# %% tags=["remove-input"]
fig




# %% tags=["remove-input"]
md(f"**Figure {i_figure}**: Annual average {layer} {vv_name} from the model {clim_range} and observations. Data is limited to the 2nd and 98th percentile of the combined model and observational data. Arrows indicate that values can exceed the colorbar limits.") 
i_figure += 1



import copy
import time
import nctoolkit as nc
import re
import importlib
import glob
import subprocess
import multiprocessing
import pathlib
import os
import pandas as pd
import string
import random
import warnings
import pickle
import xarray as xr
import oceanval.parsers as parsers
from oceanval.session import session_info
from oceanval.parsers import Validator, definitions
from multiprocessing import Manager
from tqdm import tqdm
from oceanval.utils import extension_of_directory
from oceanval.parsers import generate_mapping
from oceanval.gridded import gridded_matchup


def is_z_up(ff, variable=None):
    import netCDF4 as nc4

    try:
        ds1 = nc.open_data(ff, checks=False)
        if variable is None:
            var = ds1.variables[0]
        else:
            var = variable
        # select variable using nco
        ds1.nco_command(f"ncks -v {var}")
        ds1.run()
        ds = xr.open_dataset(ds1[0])
        ds = ds
        for x in ds[var].metpy.coordinates("longitude"):
            lon_name = x.name
        for x in ds[var].metpy.coordinates("latitude"):
            lat_name = x.name
        for x in ds[var].metpy.coordinates("time"):
            time_name = x.name
        coords = [
            x for x in list(ds.coords) if x not in [lon_name, lat_name, time_name]
        ]
        coords = [x for x in coords if "time" not in x.lower()]

        ds = nc4.Dataset(ff)
        if len(coords) == 1:

            z = ds.variables[coords[0]]
            if hasattr(z, "positive"):
                if z.positive == "down":
                    return False
                else:
                    return True
        raise ValueError(
            "Could not determine if z-axis is down from the provided file."
        )
    except:
        raise ValueError(
            "Could not determine if z-axis is down from the provided file."
        )


# a list of valid variables for validation
valid_vars = definitions.keys
# add some additionals

session_warnings = Manager().list()

nc.options(parallel=True)
nc.options(progress=False)


def mm_match(ff, model_variable, df, df_times, ds_depths, variable, df_all, layer=None):
    """
    Parameters
    -------------
    ff: str
        Path to file
    model_variable: str
        Variable name in the simulation
    df: pd.DataFrame
        Dataframe of observational data
    df_times: pd.DataFrame
        Dataframe of observational data 
    ds_depths: list
        Depths to match

    """


    if session_info["cache"]:
        try:
            ff_read = (
                session_info["cache_mapping"]
                .query("path == @ff & layer == @layer & variable == @variable")
                .output
            )
            # read this pickle file in
            with open(ff_read.values[0], "rb") as f:
                df_ff = pickle.load(f)
                df_all.append(df_ff)
                return None
        except:
            pass

    df_ff = None

    if ds_depths is not None:
        nc.session.append_safe(ds_depths[0])
    try:
        with warnings.catch_warnings(record=True) as w:
            ds = nc.open_data(ff, checks=False)
            var_match = model_variable.split("+")

            valid_locs = ["lon", "lat", "year", "month", "day", "depth"]
            valid_locs = [x for x in valid_locs if x in df.columns]

            valid_times = (
                "year" in df.columns or "month" in df.columns or "day" in df.columns
            )

            if valid_times:
                df_locs = (
                    df_times.query("path == @ff")
                    .merge(df)
                    .loc[:, valid_locs]
                    .drop_duplicates()
                    .reset_index(drop=True)
                )
            else:
                df_locs = df.loc[:, valid_locs]

            ds.subset(variables=var_match)
            #
            ds1 = nc.open_data(ds[0], checks=False)
            ds_contents = ds1.contents
            the_var = model_variable.split("+")[0]

            if list(ds_contents.query("variable == @the_var").nlevels)[0] > 1:
                if layer == "surface":
                    ds.cdo_command("topvalue")
                else:
                    if session_info["invert"]:
                        ds.invert_levels()

            if (
                "year" in df_locs.columns
                or "month" in df_locs.columns
                or "day" in df_locs.columns
            ):
                ff_indices = df_times.query("path == @ff")

                ff_indices = ff_indices.reset_index(drop=True).reset_index()
                ff_indices = ff_indices
                ff_indices = ff_indices.merge(df_locs)
                ff_indices = ff_indices["index"].values
                ff_indices = [int(x) for x in ff_indices]
                ff_indices = list(set(ff_indices))

                if len(ff_indices) == 0:
                    return None
                ds.subset(time=ff_indices)
            if session_info["as_missing"] is not None:
                ds.as_missing(session_info["as_missing"])
            ds.run()


            if len(var_match) > 1:
                ds.sum_all()
            the_dict = {"model_variable": model_variable}
            session_info["adhoc"] = copy.deepcopy(the_dict)

            if len(df_locs) > 0:
                ds.run()
                if ds_depths is not None:
                    df_ff = ds.match_points(
                        df_locs, depths=ds_depths, quiet=True, max_extrap=0
                    )
                else:
                    df_ff = ds.match_points(
                        df_locs, quiet=True, max_extrap=0
                    )
                if df_ff is not None:
                    df_ff = df_ff.dropna().reset_index(drop=True)

                if df_ff is not None:
                    valid_vars = ["lon", "lat", "year", "month", "day", "depth"]
                    for vv in ds.variables:
                        valid_vars.append(vv)
                    valid_vars = [x for x in valid_vars if x in df_ff.columns]
                    df_ff = df_ff.loc[:, valid_vars]
                    # add this to the cache if necessary
                    if session_info["cache"]:
                        cache_dir = session_info["cache_dir"]
                        if cache_dir is not None:
                            if not os.path.exists(cache_dir):
                                os.makedirs(cache_dir)
                            # create a random string
                            random_string = "".join(
                                random.choices(
                                    string.ascii_lowercase + string.digits, k=10
                                )
                            )
                            cache_file = (
                                cache_dir
                                + "output/"
                                + "/matchup_"
                                + layer
                                + "_"
                                + variable
                                + "_"
                                + random_string
                                + ".pkl"
                            )
                            if not os.path.exists(cache_dir + "/output"):
                                os.makedirs(cache_dir + "/output")
                            with open(cache_file, "wb") as f:
                                pickle.dump(df_ff, f)
                            # add a mapping to session_info["cache_files"]
                            # add to session_info["cache_dir"] + "/mappings"
                            mapping_file = (
                                cache_dir
                                + "/mappings/mapping_"
                                + layer
                                + "_"
                                + variable
                                + "_"
                                + random_string
                                + ".pkl"
                            )
                            if not os.path.exists(cache_dir + "/mappings"):
                                os.makedirs(cache_dir + "/mappings")
                            # dump ff
                            with open(mapping_file, "wb") as f:
                                output_dir = {ff: cache_file}
                                pickle.dump(output_dir, f)

                    df_all.append(df_ff)
            else:
                return None
        if df_ff is not None:
            for ww in w:
                if str(ww.message) not in session_warnings:
                    session_warnings.append(str(ww.message))

    except Exception as e:
        print(e)


def get_time_res(x, folder=None):
    """
    Get the time resolution of the netCDF files

    Parameters
    -------------
    x : str
        The extension of the file
    folder : str
        The folder containing the netCDF files

    Returns
    -------------
    res : str
        The time resolution of the netCDF files

    """

    final_extension = extension_of_directory(folder)

    if final_extension[0] == "/":
        final_extension = final_extension[1:]

    wild_card = final_extension + x
    wild_card = wild_card.replace("**", "*")
    # replace double stars with 1
    wild_card = wild_card.replace("**", "*")

    wild_card = os.path.basename(wild_card)
    for y in pathlib.Path(folder).glob(wild_card):
        path = y
        # convert to string
        path = str(path)
        break

    ds = nc.open_data(path, checks=False)
    ds_times = ds.times
    try:
        months = [x.month for x in ds_times]
        days = [x.day for x in ds_times]
        years = [x.year for x in ds_times]
    except:
        years = [int(str(x).split("T")[0].split("-")[0]) for x in ds.times]
        months = [int(str(x).split("T")[0].split("-")[1]) for x in ds.times]
        days = [int(str(x).split("T")[0].split("-")[2]) for x in ds.times]
    df_times = pd.DataFrame({"month": months, "day": days, "year": years})

    n1 = len(
        df_times.loc[:, ["month", "year"]].drop_duplicates().reset_index(drop=True)
    )
    n2 = len(df_times)
    if n1 == n2:
        return "m"
    else:
        return "d"


random_files = []


def extract_variable_mapping(folder, exclude=[], n_check=None):
    """
    Find paths to netCDF files
    Parameters
    -------------
    folder : str
        The folder containing the netCDF files
    exclude : list
        List of strings to exclude

    Returns
    -------------
    all_df : pd.DataFrame
        A DataFrame containing the paths to the netCDF files
    """

    # add restart to exclude
    exclude.append("restart")

    while True:

        levels = session_info["levels_down"]

        new_directory = folder + "/"
        if levels > 0:
            for i in range(levels + 1):
                dir_glob = glob.glob(new_directory + "/**")
                # randomize dir_glob

                random.shuffle(dir_glob)
                for x in dir_glob:
                    # figure out if the the base directory is an integer
                    try:
                        if levels != 0:
                            y = int(os.path.basename(x))
                        new_directory = x + "/"
                    except:
                        pass
        options = glob.glob(new_directory + "/**.nc")
        # if n_check is not None and an integer, limit options to n_check
        if n_check is not None and isinstance(n_check, int):
            options = random.sample(options, min(n_check, len(options)))
        if True:
            options = [x for x in options if "restart" not in os.path.basename(x)]

        if len([x for x in options if ".nc" in x]) > 0:
            break

    all_df = []
    print("********************************")
    print("Parsing model information from netCDF files")

    # remove any files from options if parts of exclude are in them
    options_thickness = copy.deepcopy(options)
    for exc in exclude:
        options = [x for x in options if f"{exc}" not in os.path.basename(x)]

    print("Searching through files in a random directory to identify variable mappings")
    # randomize options
    for ff in tqdm(options):
        random_files.append(ff)
        ds = nc.open_data(ff, checks=False)
        stop = True
        ds_dict = generate_mapping(ds)
        try:
            ds_dict = generate_mapping(ds)
            stop = False
        # output error and ff
        except:
            pass
        if stop:
            continue

        ds_vars = ds.variables

        if len([x for x in ds_dict.values() if x is not None]) > 0:
            new_name = ""
            for x in os.path.basename(ff).split("_"):
                try:
                    y = int(x)
                    if len(new_name) > 0:
                        new_name = new_name + "_**"
                    else:
                        new_name = new_name + "**"
                except:
                    if len(new_name) > 0:
                        new_name = new_name + "_" + x
                    else:
                        new_name = x
            # replace integers in new_name with **

            new_dict = dict()
            for key in ds_dict:
                if ds_dict[key] is not None:
                    #new_dict[ds_dict[key]] = [key]
                    new_dict[key] = [ds_dict[key]]
            # new_name. Replace numbers between _ with **

            # replace integers with 4 or more digits with **
            new_name = re.sub(r"\d{4,}", "**", new_name)
            # replace strings of the form _12. with _**.
            new_name = re.sub(r"\d{2,}", "**", new_name)

            all_df.append(
                pd.DataFrame.from_dict(new_dict).melt().assign(pattern=new_name)
            )
    
    try: 
        all_df = pd.concat(all_df).reset_index(drop=True)
    except:
        raise ValueError("No netCDF files found with any of the model variables.")
    #  rename variable-value, and value-variable
    all_df = all_df.rename(columns={"variable": "value", "value": "variable"}) 

    patterns = set(all_df.pattern)
    resolution_dict = dict()
    for folder in patterns:
        resolution_dict[folder] = get_time_res(folder, new_directory)
    all_df["resolution"] = [resolution_dict[x] for x in all_df.pattern]

    all_df = (
        all_df.sort_values("resolution").groupby("value").head(1).reset_index(drop=True)
    )
    all_df = all_df.rename(columns={"variable": "model_variable"})
    all_df = all_df.rename(columns={"value": "variable"})
    all_df = all_df.drop(columns="resolution")
    all_df = all_df.loc[:, ["variable", "model_variable", "pattern"]]

    return all_df


def matchup(
    sim_dir=None,
    start=None,
    end=None,
    lon_lim=None,
    lat_lim=None,
    cores=6,
    thickness=None,
    n_dirs_down=2,
    point_time_res=["year", "month", "day"],
    overwrite=True,
    ask=True,
    out_dir="",
    exclude=[],
    cache=False,
    n_check=None,
    as_missing=None,
    **kwargs,
):
    """
    Match up model with observational data

    Parameters
    -------------

    sim_dir : str
        Folder containing model output
    start : int
        Start year. First year of the simulations to matchup.
        This must be supplied
    end : int
        End year. Final year of the simulations to matchup.
        This must be supplied
    cores : int
        Number of cores to use for parallel extraction and matchups of data.
        Default is 6, or the system cores if less than 6.
        If you use a large number of cores you may run into RAM issues, so keep an eye on things.
    thickness : str
        Path to a thickness file, i.e. cell vertical thickness or the name of the thickness variable. This only needs to be supplied if the variable is missing from the raw data.
        If the cell_thickness variable is in the raw data, it will be used, and thickness does not need to be supplied.
    n_dirs_down : int
        Number of levels down to look for netCDF files. Default is 2, ie. the files are of the format */*/*.nc.
    point_time_res : list or dict
        List of strings or a dict. Default is ['year', 'month', 'day']. This is the time resolution of the point data matchup.
        If you want fine-grained control, provide a dictionary where the key is the variable and the value is a list of strings.
        If you provide this list make sure all variables have keys, or else provide a key called "default" with a value to use when the variable is not stated explicitly.
    exclude : list
        List of strings to exclude. This is useful if you have files in the directory that you do not want to include in the matchup.
    lon_lim : list
        List of two floats, which must be provided. The first is the minimum longitude, the second is the maximum longitude. Default is None.
    lat_lim : list
        List of two float. Default is None, so no spatial subsetting will occur. The first is the minimum latitude, the second is the maximum latitude. Default is None.
    ask : bool
        If True, the user will be asked if they are happy with the matchups. Default is True.
    out_dir : str
        Path to output directory. Default is "", so the output will be saved in the current directory.
    as_missing : float or list
        Value(s) to treat as missing in the model data. Default is None.
    kwargs: dict
        Additional arguments

    Returns
    -------------
    None
    Data will be stored in the matched directory.

    """
    session_info["as_missing"] = as_missing


    # store short title
    gridded = None
    point = None

    if isinstance(point_time_res, str):
        point_time_res = [point_time_res]
    if isinstance(point_time_res, list) is False:
        raise TypeError("point_time_res must be a list or a string")
    session_info["point_time_res"] = copy.deepcopy(point_time_res)
    # check it is str or list

    # make point a list if it's None
    if point is None:
        point = dict()
        point["all"] = []
        point["surface"] = []

    # if point is str, make it a list
    if isinstance(point, str):
        point = [point]

    if not isinstance(point, list) and not isinstance(point, dict):
        raise TypeError("point must be a list or a string")
    # if point is a list, convert to a dictionary
    if isinstance(point, list):
        point_new = copy.deepcopy(point)
        point = dict()
        point["all"] = point_new
        point["surface"] = []
    # loop through definition keys


    if len(definitions.keys) == 0:
        raise ValueError("You do not appear to have asked for any variables to be validated!")
    for key in definitions.keys:
        try:
            if definitions[key].vertical_point is False:
                if key not in point["surface"]:
                    point["surface"].append(key)
        except:
            pass
        try:
            if definitions[key].vertical_point is True:
                if key not in point["all"]:
                    point["all"].append(key)
        except:
            pass
        # do the same for gridded
        try:
            if definitions[key].gridded:
                if gridded is None:
                    gridded = []
                if key not in gridded:
                    gridded.append(key)
        except:
            pass
    
    if isinstance(point, dict):
        # check keys are valid
        for key in point.keys():
            if key not in ["all", "surface"]:
                raise ValueError("point dictionary keys must be 'all', 'surface'")
        # check values are lists
        for key in point.keys():
            if isinstance(point[key], str):
                point[key] = [point[key]]
            # if it's None, convert to empty list
            if point[key] is None:
                point[key] = []
            if not isinstance(point[key], list):
                raise TypeError("point dictionary values must be lists or strings")
        # if any keys are absent, make them empty lists
        for key in ["all", "surface"]:
            if key not in point.keys():
                point[key] = []

    # if gridded is str, make it a list
    if isinstance(gridded, str):
        gridded = [gridded]
    # None too
    if gridded is None:
        gridded = []

    ignore_invert_check = False
    # go through kwargs
    for key, value in kwargs.items():
        if key == "ignore_invert_check":
            ignore_invert_check = value
            # must be a bool
            if not isinstance(ignore_invert_check, bool):
                raise TypeError("ignore_invert_check must be a boolean")

    # convert sim_dir to hard path
    if sim_dir is not None:
        sim_dir = os.path.abspath(sim_dir)

    # if cache is True, create a cache directory in out_dir
    out_dir = os.path.expanduser(out_dir)
    if cache:
        if out_dir == "":
            out_dir = "./"
        cache_dir = out_dir + "/.cache_oceanval/"
        if not os.path.exists(cache_dir):
            os.makedirs(cache_dir)
        session_info["cache_dir"] = cache_dir
        session_info["cache"] = True
        # create a mappings directory in cache_dir
        mappings_dir = cache_dir + "/mappings/"
        if not os.path.exists(mappings_dir):
            os.makedirs(mappings_dir)
        # list files in mappings_dir
        paths = glob.glob(mappings_dir + "/*.pkl")
        mapping_df = []
        for ff in paths:
            with open(ff, "rb") as f:
                output_dir = pickle.load(f)
            ff_path = list(output_dir.keys())[0]
            ff_output = list(output_dir.values())[0]
            ff_layer = os.path.basename(ff).split("_")[1]
            ff_variable = os.path.basename(ff).split("_")[2]
            mapping_df.append(
                pd.DataFrame(
                    {
                        "variable": [ff_variable],
                        "layer": [ff_layer],
                        "path": [ff_path],
                        "output": [ff_output],
                    }
                )
            )
        if len(mapping_df) > 0:
            mapping_df = pd.concat(mapping_df).reset_index(drop=True)
        session_info["cache_mapping"] = mapping_df

    else:
        session_info["cache_dir"] = None
        session_info["cache"] = False

    if not isinstance(ask, bool):
        raise TypeError("ask must be a boolean")
    if not isinstance(overwrite, bool):
        raise TypeError("overwrite must be a boolean")
    # check n_dirs_down is an integer
    if not isinstance(n_dirs_down, int):
        raise TypeError("n_dirs_down must be an integer")
    if n_dirs_down < 0:
        raise ValueError("n_dirs_down must be a positive integer")
    # check if exclude is a list or str
    if not isinstance(exclude, list):
        if isinstance(exclude, str):
            exclude = [exclude]
        else:
            raise TypeError("exclude must be a list or a string")

    if thickness is not None:
        if isinstance(thickness, str):
            # if it ends with .nc check it exists
            if thickness.endswith(".nc"):
                if not os.path.exists(thickness):
                    raise FileNotFoundError(f"{thickness} does not exist")
    if thickness == "z-level":
        thickness = "z_level"
    if thickness == "z level":
        thickness = "z_level"
    if thickness == "z_level":
        session_info["z_level"] = True
    else:
        session_info["z_level"] = False

    if cores == 6:
        if cores > os.cpu_count():
            cores = os.cpu_count()
            print(
                f"Setting cores to {cores} as this is the number of cores available on your system"
            )
    nc.options(cores=cores)
    session_info["cores"] = cores

    # check lon_lim and lat_lim are lists

    if lon_lim is None:
        if lat_lim is not None:
            raise TypeError("lat_lim must be a list")
    if lat_lim is None:
        if lon_lim is not None:
            raise TypeError("lon_lim must be a list")

    # add this info to session_info
    session_info["lon_lim"] = lon_lim
    session_info["lat_lim"] = lat_lim

    ds_depths = None

    if start is None:
        raise ValueError("Please provide a start year")

    if end is None:
        raise ValueError("Please provide an end year")

    if isinstance(start, int) is False:
        raise TypeError("Start must be an integer")

    if isinstance(end, int) is False:
        raise TypeError("End must be an integer")

    # check if the sim_dir exists
    if sim_dir is None:
        raise ValueError("Please provide a sim_dir directory")

    if not os.path.exists(sim_dir):
        raise ValueError(f"{sim_dir} does not exist")

    # set up session info, which will be needed by gridded_matchup
    session_info["overwrite"] = overwrite

    # add out_dir to session_info
    if out_dir != "":
        session_info["out_dir"] = out_dir + "/"
    else:
        session_info["out_dir"] = ""

    ff = session_info["out_dir"] + "oceanval_matchups/short_titles.pkl"
    if os.path.exists(ff):
        with open(ff, "rb") as f:
            short_titles = pickle.load(f)
    else:
        short_titles = dict()
    session_info["short_title"] = short_titles | session_info["short_title"]

    # raise ValueError(session_info["short_title"])

    if n_dirs_down is not None:
        session_info["levels_down"] = n_dirs_down
    else:
        session_info["levels_down"] = 2

    sim_start = -1000
    sim_end = 10000
    for key in kwargs:
        key_failed = True
        if key[:3] == "fol":
            if sim_dir is None:
                sim_dir = kwargs[key]
                key_failed = False
        if key_failed:
            raise ValueError(f"{key} is not a valid argument")

    if end is not None:
        sim_end = end

    if start is not None:
        sim_start = start

    # check validity of variables chosen

    all_df = None

    var_choice = gridded
    var_choice = list(set(var_choice))
    if isinstance(gridded, str):
        var_choice = [gridded]

    for vv in var_choice:
        if vv not in valid_vars and vv != "all":
            # suggest another variable based on similarity to valid_vars
            from difflib import get_close_matches

            close = get_close_matches(vv, valid_vars)
            if len(close) > 0:
                raise ValueError(
                    f"{vv} is not a valid variable. Did you mean {close[0]}?"
                )
            if vv != "default":
                raise ValueError(
                    f"{vv} is not a valid variable. Please choose from {valid_vars}"
                )

    all_df = extract_variable_mapping(sim_dir, exclude=exclude, n_check=n_check)
    # check if any model_variable is None
    var_found = list(all_df.variable.unique())
    missing = ",".join([x for x in var_choice if x not in var_found])
    # 
    if len(missing) > 0:
        error = f"The model variables specified do not appear to be in the simulation output for the following: {missing}. Please check the model_variable names and try again."
        raise ValueError(error)

    # add in anything that is missing
    all_vars = valid_vars

    missing_df = pd.DataFrame({"variable": all_vars}).assign(
        model_variable=None, pattern=None
    )

    all_df = (
        pd.concat([all_df, missing_df])
        .groupby("variable")
        .head(1)
        .reset_index(drop=True)
        )
    # check if the variables are in all_df

    pattern = all_df.reset_index(drop=True).iloc[0, :].pattern

    final_extension = extension_of_directory(sim_dir)

    if final_extension[0] == "/":
        final_extension = final_extension[1:]

    wild_card = final_extension + pattern
    wild_card = wild_card.replace("**", "*")
    for x in pathlib.Path(sim_dir).glob(wild_card):
        path = x
        # convert to string
        path = str(path)
        break

    try:
        ds = nc.open_data(path, checks=False)
    except:
        raise ValueError("Problems finding files. Check n_dirs_down arg")

    # check length of lon_lim and lat_lim
    if session_info["lon_lim"] is not None:
        if len(session_info["lon_lim"]) != 2 or len(session_info["lat_lim"]) != 2:
            raise ValueError(
                "lon_lim and lat_lim must be lists of two floats, e.g. [-18, 9] and [42, 63]"
            )


    vars_available = list(
        all_df
        # drop rows where pattern is None
        .dropna()
        # get all variables
        .variable
    )
    # check variables chosen are valid

    remove = []

    gridded = [x for x in gridded if x in vars_available]
    for key in point.keys():
        point[key] = [x for x in point[key] if x in vars_available]

    for vv in point["all"]:
        if definitions[vv].vertical_point is False:
            point["all"].remove(vv)
            point["surface"].append(vv)

    var_chosen = gridded + point["all"] + point["surface"]
    var_chosen = list(set(var_chosen))

    # create oceanval_matchups directory
    if not os.path.exists("oceanval_matchups"):
        if session_info["out_dir"] != "":
            # recusively create the directory
            os.makedirs(session_info["out_dir"] + "/oceanval_matchups", exist_ok=True)
        else:
            os.mkdir("oceanval_matchups")

    invert_thickness = False
    point_all = point["all"] + point["surface"]

    # go through variables in definitions
    thick_check = False
    for vv in var_chosen:
        # identifical if vertical_gridded is True
        try:
            if definitions[vv].vertical_gridded:
                thick_check = True
        except:
            pass
    if len(point["all"]) > 0:
        thick_check = True

    if session_info["z_level"]:
        thick_check = False

    if thick_check:
        print("Sorting out thickness")
        ds_depths = False
        with warnings.catch_warnings(record=True) as w:
            # extract the thickness dataset
            cell_thickness_found = False
            if thickness is not None and os.path.exists(thickness):
                ds_thickness = nc.open_data(thickness, checks=False)
                invert_thickness = is_z_up(ds_thickness[0])
                if len(ds_thickness.variables) != 1:
                    if (
                        len(
                            [x for x in ds_thickness.variables if "cell_thickness" in x]
                        )
                        == 0
                    ):
                        raise ValueError(
                            "The thickness file has more than one variable and none include cell_thickness. Please provide a single variable!"
                        )
                ds_thickness.rename({ds_thickness.variables[0]: "cell_thickness"})
                cell_thickness_found = True
                thickness = "cell_thickness"
            else:
                print(
                    "Vertical thickness is required for your matchups, but they are not supplied"
                )
                print("Searching through simulation output to find it")
                if thickness is None:
                    raise ValueError(
                        "Please provide the name of the thickness variable"
                    )
                for ff in random_files:
                    print("Checking file for thickness: " + ff)
                    # do this quietly
                    with warnings.catch_warnings(record=True) as w:
                        ds_thickness = nc.open_data(ff, checks=False)
                        if thickness in ds_thickness.variables:
                            cell_thickness_found = True
                            invert_thickness = is_z_up(ff, thickness)
                            break
                        else:
                            if (
                                len(
                                    [
                                        x
                                        for x in ds_thickness.variables
                                        if thickness in x
                                    ]
                                )
                                > 0
                            ):
                                cell_thickness_found = True
                                invert_thickness = is_z_up(ff, thickness)
                                break

            if not cell_thickness_found:
                raise ValueError("Unable to find cell_thickness")

            if os.path.exists(thickness) == False:
                if len(ds_thickness.times) > 0:
                    ds_thickness.subset(time=0, variables=f"{thickness}*")
                else:
                    ds_thickness.subset(variables=f"{thickness}*")
            ds_thickness.run()
            var_sel = (
                ds_thickness.contents.query(f"variable.str.contains('{thickness}')")
                .query("nlevels > 1")
                .variable
            )
            ds_thickness.subset(variables=var_sel)
            if session_info["as_missing"] is not None:
                ds_thickness.as_missing(session_info["as_missing"])
            if len(ds_thickness.variables) > 1:
                if "cell_thickness" in ds_thickness.variables:
                    ds_thickness.subset(variables=f"{thickness}*")
                else:
                    ds_thickness.subset(variables=ds_thickness.variables[0])
            ds_thickness.run()
            print(f"Thickness variable is {ds_thickness.variables[0]} from {ff}")
            #####
            # now output the bathymetry if it does not exists

            ff_bath = session_info["out_dir"] + "oceanval_matchups/model_bathymetry.nc"
            if not os.path.exists(ff_bath):
                ds_bath = ds_thickness.copy()
                ds_bath.vertical_sum()
                ds_bath.to_nc(ff_bath, zip=True)

            if invert_thickness and ignore_invert_check is False:
                # user check
                x = input(
                    "The thickness data appears to have the sea surface at the bottom (i.e. increasing depth values down. DO NOT PROCEED IF THIS IS A NEMO SIMULATION. Is this correct? (y/n) "
                )
                if x.lower() == "n":
                    return None

            ds_thickness.run()
            if invert_thickness:
                ds_thickness.invert_levels()
                ds_thickness.run()
            ds_thickness_sim = ds_thickness.copy()

            ds_depths = ds_thickness.copy()

            ds_depths.vertical_cumsum()
            ds_thickness / 2
            ds_depths - ds_thickness
            ds_depths.run()
            ds_depths.rename({ds_depths.variables[0]: "depth"})
            ds_depths.run()

        for ww in w:
            if str(ww.message) not in session_warnings:
                session_warnings.append(str(ww.message))
        if ds_depths is False:
            raise ValueError(
                "You have asked for variables that require the specification of thickness"
            )
        print("Thickness is sorted out")

        ds_depths.run()
    if ds_depths is not None:
        session_info["ds_depths"] = ds_depths[0]
        session_info["ds_thickness"] = ds_thickness_sim[0]
    else:
        session_info["ds_depths"] = thickness
        session_info["ds_thickness"] = None 

    session_info["invert"] = invert_thickness

    # add the global checker here
    # sort all_df alphabetically by variable
    all_df = all_df.sort_values("variable").reset_index(drop=True)
    gridded.sort()
    print("Variables that will be matched up")
    print("******************************")
    if len(gridded) > 0:
        print(
            f"The following variables will be matched up with gridded surface data: {','.join(gridded)}"
        )

    print("******************************")
    print(f"** Inferred mapping of model variable names from {sim_dir}")

    all_df_print = copy.deepcopy(all_df).reset_index(drop=True)

    # new tidied variable
    new_variable = []
    for i in range(len(all_df_print)):
        if all_df.variable[i] in var_chosen:
            if all_df.pattern[i] is not None:
                new_variable.append(all_df.variable[i] + "**")
            else:
                new_variable.append(all_df.variable[i])
        else:
            new_variable.append(all_df.variable[i])
    all_df_print["variable"] = new_variable
    all_df_print = all_df_print.sort_values("variable")
    # move variable with ** in it to the top of the data frame
    all_df_print = all_df_print.sort_values(
        by="variable", key=lambda x: x.str.replace(r".*\*\*", "", regex=True)
    )
    all_df_print = all_df_print.reset_index(drop=True)
    # add a new row with --- in it after variables with ** in them
    new_row = pd.DataFrame(
        {"variable": ["---"], "model_variable": ["---"], "pattern": ["---"]}
    )
    all_df_print = pd.concat([all_df_print, new_row], ignore_index=True)
    all_df_print = all_df_print.sort_values(
        by="variable", key=lambda x: x.str.replace(r".*\*\*", "", regex=True)
    )
    all_df_print = all_df_print.reset_index(drop=True)

    print(all_df_print.to_string(index=False))

    print("Are you happy with these matchups? Y/N")

    if ask:
        x = input()
    else:
        x = "y"

    if x.lower() not in ["y", "n"]:
        print("Provide Y or N")
        x = input()

    if x.lower() == "n":
        print("Please adjust your variable names and try again")
        return None

    if session_info["out_dir"] != "":
        out = session_info["out_dir"] + "/oceanval_matchups/mapping.csv"
    else:
        out = "oceanval_matchups/mapping.csv"
    # check directory exists for out
    out_folder = os.path.dirname(out)
    if not os.path.exists(out_folder):
        os.makedirs(out_folder)

    df_out = all_df.dropna().reset_index(drop=True)
    final_extension = extension_of_directory(sim_dir)
    df_out["pattern"] = [sim_dir + final_extension + x for x in df_out.pattern]
    df_out.to_csv(out, index=False)
    # restrict all_df to only variables chosen
    all_df = all_df.query("variable in @var_chosen").reset_index(drop=True)

    final_extension = extension_of_directory(sim_dir)

    # combine all variables into a list
    all_vars = gridded + point["all"] + point["surface"]
    all_vars = list(set(all_vars))

    df_variables = all_df.query("variable in @all_vars").reset_index(drop=True)
    # remove rows where model_variable is None
    df_variables = df_variables.dropna().reset_index(drop=True)

    patterns = list(set(df_variables.pattern))

    times_dict = dict()

    print("*************************************")

    if ds_depths is not None:
        # run cdo griddes on the file and figure out if generic shows up

        output = subprocess.check_output(["cdo", "griddes", ds_depths[0]])
        if "generic" in str(output):
            ff1 = path
            ds_depths.cdo_command(f"setgrid,{ff1}")

    ff = session_info["out_dir"] + "oceanval_matchups/times_dict.pkl"
    # read this in
    try:
        with open(ff, "rb") as f:
            times_dict = pickle.load(f)
    except:
        times_dict = dict()
        pass

    print("*************************************")
    for pattern in patterns:
        print(f"Indexing file time information for {pattern} files")
        final_extension = extension_of_directory(sim_dir)
        ensemble = glob.glob(sim_dir + final_extension + pattern)
        for exc in exclude:
            ensemble = [x for x in ensemble if f"{exc}" not in os.path.basename(x)]

        try:
            ds = xr.open_dataset(ensemble[0])
            time_name = [x for x in list(ds.dims) if "time" in x][0]
        except:
            ds = xr.open_dataset(ensemble[0], decode_times=False)
            time_name = [x for x in list(ds.dims) if "time" in x][0]

        for ff in tqdm(ensemble):
            if ff in times_dict:
                continue
            if "restart" in ff:
                continue

            try:
                ds = xr.open_dataset(ff)
                ff_month = [int(x.dt.month) for x in ds[time_name]]
                ff_year = [int(x.dt.year) for x in ds[time_name]]
                days = [int(x.dt.day) for x in ds[time_name]]
            except:
                ds = nc.open_data(ff, checks=False)
                ds_times = ds.times
                ff_month = [int(x.month) for x in ds_times]
                ff_year = [int(x.year) for x in ds_times]
                days = [int(x.day) for x in ds_times]

            df_ff = pd.DataFrame(
                {
                    "year": ff_year,
                    "month": ff_month,
                    "day": days,
                }
            )
            times_dict[ff] = df_ff


    # save this as a pickle
    with open(session_info["out_dir"] + "oceanval_matchups/times_dict.pkl", "wb") as f:
        pickle.dump(times_dict, f)

    # figure out the lon/lat extent in the model
    with warnings.catch_warnings(record=True) as w:
        lons = session_info["lon_lim"]
        lats = session_info["lat_lim"]

    all_df = all_df.dropna().reset_index(drop=True)
    df_mapping = all_df
    good_model_vars = [x for x in all_df.model_variable if x is not None]

    df_mapping = all_df

    point_all = point["all"] + point["surface"]
    if len(point_all) > 0:
        print("********************************")
        print("Matching up with observational point data")
        print("********************************")

        # if model_variable is None remove from all_df

        for key, value in point.items():
            the_vars = list(df_mapping.dropna().variable)
            var_choice = [x for x in var_choice if x in the_vars]
            point_vars = value
            depths = copy.deepcopy(key)
            layer = depths

            # sort the list
            point_vars.sort()

            for vv in point_vars:


                # try finding source in definitions
                variable = vv
                source = definitions[variable].point_source

                if session_info["out_dir"] != "":
                    out = f"{session_info['out_dir']}/oceanval_matchups/point/{layer}/{variable}/{source}/{source}_{layer}_{variable}.csv"
                else:
                    out = f"oceanval_matchups/point/{layer}/{variable}/{source}/{source}_{layer}_{variable}.csv"

                if os.path.exists(out) and not overwrite:
                    continue

                all_df = df_mapping
                all_df = all_df.query("model_variable in @good_model_vars").reset_index(
                    drop=True
                )

                all_df = all_df.dropna()
                all_df = all_df.query("variable == @vv").reset_index(drop=True)
                patterns = list(set(all_df.pattern))

                for pattern in patterns:
                    final_extension = extension_of_directory(sim_dir)
                    ensemble = glob.glob(sim_dir + final_extension + pattern)
                    for exc in exclude:
                        ensemble = [
                            x for x in ensemble if f"{exc}" not in os.path.basename(x)
                        ]

                    df_times = []
                    days = []
                    for ff in ensemble:
                        df_ff = times_dict[ff]
                        df_times.append(
                            pd.DataFrame(
                                {
                                    "month": df_ff.month,
                                    "year": df_ff.year,
                                    "day": df_ff.day,
                                }
                            ).assign(path=ff)
                        )
                    df_times = pd.concat(df_times)

                    # Idea: figure out if it is monthly or daily data
                    # This might help speed things up

                    sim_paths = list(set(df_times.path))
                    sim_paths.sort()
                    # write to the report

                    min_year = df_times.year.min()
                    max_year = df_times.year.max()
                    session_info["min_year"] = min_year
                    # factor in start
                    session_info["min_year"] = max(session_info["min_year"], sim_start)
                    session_info["max_year"] = max_year
                    # factor in end
                    session_info["max_year"] = min(session_info["max_year"], sim_end)

                    def point_match(
                        variable, layer="all", ds_depths=None, df_times=None
                    ):
                        with warnings.catch_warnings(record=True) as w:
                            point_variable = variable
                            model_variable = list(
                                all_df.query(
                                    "variable == @point_variable"
                                ).model_variable
                            )[0]

                            paths = glob.glob(
                                f"{definitions[variable].point_dir}/**.csv"
                            )

                            # try finding source in definitions
                            source = definitions[variable].point_source

                            if session_info["out_dir"] != "":
                                out = f"{session_info['out_dir']}/oceanval_matchups/point/{layer}/{variable}/{source}/{source}_{layer}_{variable}.csv"
                            else:
                                out = f"oceanval_matchups/point/{layer}/{variable}/{source}/{source}_{layer}_{variable}.csv"

                            for exc in exclude:
                                paths = [
                                    x
                                    for x in paths
                                    if f"{exc}" not in os.path.basename(x)
                                ]
                            
                            def read_csv_simyears(ff, layer = None):
                                df = pd.read_csv(ff)
                                min_year = session_info["min_year"]
                                max_year = session_info["max_year"]
                                if "year" in df.columns:
                                    df = df.query(
                                        "year >= @min_year and year <= @max_year"
                                    ).reset_index(drop=True)
                                if layer == "surface":
                                    if "depth" in df.columns:
                                        df = df.query("depth <= 5").reset_index(
                                            drop=True
                                        )
                                        # drop depth
                                return df

                            df = pd.concat([read_csv_simyears(x, layer) for x in paths])
                            if "year" in df.columns:
                                # find point_start
                                point_start = definitions[variable].point_start
                                point_end = definitions[variable].point_end
                                df = df.query(
                                    "year >= @point_start and year <= @point_end"
                                ).reset_index(drop=True)

                            # remove source if it's in df
                            if "source" in df.columns:
                                df = df.query("source == @source").reset_index(
                                    drop=True
                                )
                            # if it exists, coerce year to int
                            if "year" in df.columns:
                                df = df.assign(year=lambda x: x.year.astype(int))
                                # subset to
                            if "month" in df.columns:
                                df = df.assign(month=lambda x: x.month.astype(int))
                            if "day" in df.columns:
                                df = df.assign(day=lambda x: x.day.astype(int))
                            if layer == "surface":
                                if "depth" in df.columns:
                                    df = df.query("depth <= 5").reset_index(drop=True)
                                    # drop depth
                                    df = df.drop(columns=["depth"])

                            # extract point_time_res from dictionary
                            point_time_res = copy.deepcopy(
                                session_info["point_time_res"]
                            )
                            for x in [
                                x
                                for x in ["year", "month", "day"]
                                if x not in point_time_res
                            ]:
                                if x in df.columns:
                                    df = df.drop(columns=x)

                            sel_these = point_time_res
                            sel_these = [x for x in df.columns if x in sel_these]
                            if "year" in df.columns:
                                paths = list(
                                    set(
                                        df.loc[:, sel_these]
                                        .drop_duplicates()
                                        .merge(df_times)
                                        .path
                                    )
                                )
                            else:
                                paths = list(set(df_times.path))

                            if len(paths) == 0:
                                print(f"No matching times for {variable}")

                            manager = Manager()

                            df_times_new = copy.deepcopy(df_times)


                            valid_cols = [
                                "lon",
                                "lat",
                                "day",
                                "month",
                                "year",
                                "depth",
                                "observation",
                            ]
                            select_these = [x for x in df.columns if x in valid_cols]

                            if len(df) == 0:
                                print("No data for this variable")
                                return None

                            if "year" not in df.columns:
                                try:
                                    point_time_res.remove("year")
                                except:
                                    pass
                            if "month" not in df.columns:
                                try:
                                    point_time_res.remove("month")
                                except:
                                    pass
                            if "day" not in df.columns:
                                try:
                                    point_time_res.remove("day")
                                except:
                                    pass

                            df_all = manager.list()

                            grid_setup = False
                            pool = multiprocessing.Pool(cores)

                            pbar = tqdm(total=len(paths), position=0, leave=True)
                            results = dict()

                            for ff in paths:

                                temp = pool.apply_async(
                                    mm_match,
                                    [
                                        ff,
                                        model_variable,
                                        df,
                                        df_times_new,
                                        ds_depths,
                                        point_variable,
                                        df_all,
                                        layer,
                                    ],
                                )

                                results[ff] = temp

                            for k, v in results.items():
                                value = v.get()
                                pbar.update(1)

                            df_all = list(df_all)
                            df_all = [x for x in df_all if x is not None]
                            # do nothing when there is no data
                            if len(df_all) == 0:
                                print(f"No data for {variable}")
                                time.sleep(1)
                                return False

                            df_all = pd.concat(df_all)

                            change_this = [
                                x
                                for x in df_all.columns
                                if x
                                not in [
                                    "lon",
                                    "lat",
                                    "year",
                                    "month",
                                    "day",
                                    "depth",
                                    "observation",
                                ]
                            ][0]
                            #
                            df_all = df_all.rename(
                                columns={change_this: "model"}
                            ).merge(df)
                            # add model to name column names with frac in them
                            df_all = df_all.dropna().reset_index(drop=True)
                            # fix the observations based on obs_unit_multiplier
                            multiplier = definitions[variable].obs_multiplier_point
                            if multiplier != 1:
                                df_all = df_all.assign(
                                    observation=lambda x: x.observation * multiplier
                                )
                            adder = definitions[variable].obs_adder_point
                            if adder != 0:
                                df_all = df_all.assign(
                                    observation=lambda x: x.observation + adder
                                )

                            grouping = copy.deepcopy(point_time_res)
                            grouping.append("lon")
                            grouping.append("lat")
                            grouping.append("depth")
                            grouping = [x for x in grouping if x in df_all.columns]
                            grouping = list(set(grouping))
                            df_all = df_all.dropna().reset_index(drop=True)
                            df_all = df_all.groupby(grouping).mean().reset_index()


                            # create directory for out if it does not exists
                            if not os.path.exists(os.path.dirname(out)):
                                os.makedirs(os.path.dirname(out))
                            if lon_lim is not None:
                                df_all = df_all.query(
                                    f"lon > {lon_lim[0]} and lon < {lon_lim[1]}"
                                )
                            if lat_lim is not None:
                                df_all = df_all.query(
                                    f"lat > {lat_lim[0]} and lat < {lat_lim[1]}"
                                )

                            if len(df_all) > 0:

                                if "year" not in point_time_res:
                                    try:
                                        df_all = df_all.drop(columns="year")
                                    except:
                                        pass
                                if "day" not in point_time_res:
                                    try:
                                        df_all = df_all.drop(columns="day")
                                    except:
                                        pass
                                if "month" not in point_time_res:
                                    try:
                                        df_all = df_all.drop(columns="month")
                                    except:
                                        pass
                                df_all.to_csv(out, index=False)
                                # save the definitions
                                out_definitions = out.replace(
                                    ".csv", "_definitions.pkl"
                                )
                                import dill
                                # get the model unit
                                ds = nc.open_data(paths[0], checks=False)
                                model_unit = list(
                                    ds.contents.query(
                                        "variable == @model_variable"
                                    ).unit
                                )[0]
                                definitions[variable].model_unit = model_unit
                                dill.dump( definitions, file=open(out_definitions, "wb"))

                                out1 = out.replace(
                                    os.path.basename(out), "matchup_dict.pkl"
                                )
                                # read in the adhoc dict in mm_match

                                point_start = -5000
                                point_end = 10000
                                try:
                                    point_start = definitions[variable].point_start
                                    point_end = definitions[variable].point_end
                                except:
                                    pass

                                min_year = session_info["min_year"]
                                max_year = session_info["max_year"]

                                if point_start > min_year:
                                    min_year = point_start
                                if point_end < max_year:
                                    max_year = point_end

                                the_dict = {
                                    "start": min_year,
                                    "end": max_year,
                                    "point_time_res": point_time_res,
                                    "model_variable": model_variable,
                                }
                                # remove the adhoc dict
                                # write to pickle
                                with open(out1, "wb") as f:
                                    pickle.dump(the_dict, f)

                                return None
                            else:
                                print(f"No data for {variable}")
                                time.sleep(1)
                                return False

                    vv_variable = definitions[vv].long_name

                    if session_info["out_dir"] != "":
                        out = glob.glob(
                            session_info["out_dir"]
                            + "/"
                            + f"oceanval_matchups/point/all/{vv}/**_all_{vv}.csv"
                        )

                    else:
                        out = glob.glob(f"oceanval_matchups/point/all/{vv}/**_all_{vv}.csv")
                    if len(out) > 0:
                        if session_info["overwrite"] is False:
                            continue

                    print(
                        f"Matching up model output of {key} {vv_variable} with in-situ observational data"
                    )

                    # try:
                    if True:
                        point_match(
                            vv, ds_depths=ds_depths, df_times=df_times, layer=key
                        )
                    # except:
                    #     pass

                    output_warnings = []
                    for ww in session_warnings:
                        if ww is not None:
                            if ww in output_warnings:
                                continue
                            if "CDO found more than one time variable" in ww:
                                continue
                            if "coordinates variable time" in ww:
                                continue
                            output_warnings.append(str(ww))

                    if len(output_warnings) > 0:
                        output_warnings = list(set(output_warnings))
                        print(f"Warnings for {vv_variable}")
                        for ww in output_warnings:
                            warnings.warn(message=ww)
                    # empty session warnings
        while len(session_warnings) > 0:
            session_warnings.pop()

    gridded_matchup(
        df_mapping=df_mapping,
        folder=sim_dir,
        var_choice=gridded,
        exclude=exclude,
        sim_start=sim_start,
        sim_end=sim_end,
        lon_lim=lon_lim,
        lat_lim=lat_lim,
        times_dict=times_dict,
    )

    if len(session_info["end_messages"]) > 0:
        print("########################################")
        print("########################################")
        print("Important messages about matchups:")
        print("*" * 30)
        # write this info to a md report

        for x in session_info["end_messages"]:
            print(x)
        print("########################################")
        print("########################################")

    # store definitions as a pickle
    ff = session_info["out_dir"] + "oceanval_matchups/definitions.pkl"
    if os.path.exists(ff):
        os.remove(ff)
    import dill

    dill.dump(definitions, file=open(ff, "wb"))

    # output short titles
    ff = session_info["out_dir"] + "oceanval_matchups/short_titles.pkl"
    short_titles = session_info["short_title"]
    with open(ff, "wb") as f:
        pickle.dump(short_titles, f)
    
    # now add to the list of variables matched up
    ff = session_info["out_dir"] + "oceanval_matchups/variables_matched.pkl"
    if os.path.exists(ff):
        with open(ff, "rb") as f:
            variables_matched = pickle.load(f)
    else:
        variables_matched = []
    for vv in list(df_mapping.variable):
        if vv not in variables_matched:
            variables_matched.append(vv)
    with open(ff, "wb") as f:
        pickle.dump(variables_matched, f)





# %% tags=["remove-input"]
md(f"## Assessing model bias for {layer} {vv_name}")
#
# A critical metric for model performance is the bias between model and observed values. Here the bias is calculated as the mean difference between model and observed values. A positive bias indicates that the model overestimates the observed values, while a negative bias indicates that the model underestimates the observed values. 



# %% tags=["remove-input"]
ds_bias = ds_model.copy()
ds_bias - ds_obs
ds_bias.tmean()
ds_bias.run()
# model plot

df_bias = ds_bias.to_dataframe().reset_index()
if "model" == ds_bias.variables[0]:
    ds_bias.set_longnames({"model": Variable + " bias"})
else:
    ds_bias.set_longnames({variable: Variable + " bias"})

# get th
ds_ave = ds_bias.copy()
ds_ave.spatial_mean()
ds_ave.rename({ds_ave.variables[0]: "bias"})
ave_bias = ds_ave.to_dataframe().bias.values[0]
ds_summary = ds_bias.copy()
ds_summary > 0 
ds_summary.spatial_mean()
ds_summary.rename({ds_summary.variables[0]: "bias"})
positive_bias = ds_summary.to_dataframe().bias.values[0] * 100
# as a percentage to 1 dp
positive_bias = round(positive_bias, 1)
the_unit = ds_summary.contents.unit.values[0]

md(f"Figure {i_figure} shows the average bias of {layer} {vv_name} simulated by the model. A positive bias indicates that the model overestimates the observation, while a negative bias indicates that the model overpredicts the observation.") 
if positive_bias > 50:
    positive_bias = str(positive_bias) + "%"
    md(f"The spatial average bias of {layer} {vv_name} is {ave_bias:.2f} {the_unit}. Overall, the model overestimates the observations in {positive_bias} of the model domain.")
else:
    negative_bias = str(100-positive_bias) + "%"
    md(f"The spatial average bias of {layer} {vv_name} is {ave_bias:.2f} {the_unit}. Overall, the model underestimates the observations in {negative_bias} of the model domain.")



# %% tags=["remove-input"]

if fix_grid:
    ds_bias_1 = ds_bias.copy()
    ds_bias_1.to_latlon(lon = [lon_min , lon_max], lat = [lat_min, lat_max], res = [lon_res, lat_res])
    # change the name
    ds_bias_1.set_longnames({ds_bias_1.variables[0]: "Model bias"})
    ds_bias_1.pub_plot(robust = True)
else:
    # change the name
    ds_bias.set_longnames({ds_bias.variables[0]: "Model bias"})
    ds_bias.pub_plot(robust = True)


# %% tags=["remove-input"]
md(f"**Figure {i_figure}**: Bias of {layer} {vv_name} from the model. A positive bias indicates that the model overestimates the observation.  For clarity, the colorbar is limited to the 2nd and 98th percentile of the data.")
i_figure += 1





import re
import glob
import os
import pandas as pd
from oceanval.parsers import Validator
definitions = Validator()
tidy_info = {}


def fix_unit(x):
    x = x.replace("/m^3", "m<sup>-3</sup>") 
    x = x.replace("/m**3", "m<sup>-3</sup>") 
    x = x.replace("/m3", "m<sup>-3</sup>")
    x = x.replace("m-3", "m<sup>-3</sup>")
    x = x.replace("m-1", "m<sup>-1</sup>")
    x = x.replace("m-2", "m<sup>-2</sup>")
    x = x.replace("/m^2", "m<sup>-2</sup>")
    x = x.replace("/m2", "m<sup>-2</sup>")
    x = x.replace("m2", "m<sup>2</sup>")
    x = x.replace("m3", "m<sup>3</sup>")
    #O_2
    x = x.replace("O_2", "O<sub>2</sub>")
    # fix CO2
    x = x.replace("CO2", "CO<sub>2</sub>")
    # fix /yr
    x = x.replace("/yr", "year<sup>-1</sup>")
    # degC
    x = x.replace("degC", "°C")
    x = x.replace("degrees C", "°C")


    return x



def df_display(df):
    # only 2 decimal places
    for col in df.columns:
        if "Number" in col:
            try:
                df[col] = df[col].astype(int)
                # put in commas
                df[col] = df[col].apply(lambda x: "{:,}".format(x))
            except:
                pass
    # if rmsd is a column title change to RMSD
    if "rmsd" in df.columns:
        df = df.rename(columns={"rmsd": "RMSD"})
    # round to 2 decimal places
    df = df.round(2)


    # coerce numeric columns to str
    # if number is in the column title, make sure the variable is int

    df = df.applymap(lambda x: f"{x:.2f}" if isinstance(x, float) else x)
    # capitalize unit column name, if it exists
    if "unit" in df.columns:
        df = df.rename(columns={"unit": "Unit"})
    # capitalize variable column name, if it exists
    if "variable" in df.columns:
        df = df.rename(columns={"variable": "Variable"})
        # fix variable names
        # capitalize variable
        df["Variable"] = df["Variable"].str.capitalize()
        # ensure "Poc " is "POC "
        # ensure "Doc" is "DOC"
    if "Variable" in df.columns:
        df["Variable"] = df["Variable"].str.replace("pCO2", "pCO<sub>2</sub>")
        df["Variable"] = df["Variable"].str.replace("CO2", "CO<sub>2</sub>")
    # if R2 is in the column name, make sure the 2 is superscript
    if "R2" in df.columns:
        df = df.rename(columns={"R2": "R²"})
    # convert nan to N/A
    df = df.replace("nan", "N/A")
    if "Region" in df.columns:
        if "Full Domain" in df["Region"].values:
            # ensure the Full Domain region is the first row
            # get the index of the row
            i_domain = df[df["Region"] == "Full Domain"].index[0]
            df1 = df.iloc[i_domain:i_domain+1]
            df2 = df.drop(i_domain)
            df = pd.concat([df1, df2])
            df = df.reset_index(drop = True)

    if len(df.columns) > 7:
        # remove concentration from variable
        df = df.assign(Variable = df["Variable"].str.replace(" concentration", ""))

    if "Unit" in df.columns:
        #format this appropriately. Markdown, superscripts etc.
        df["Unit"] = df["Unit"].apply(fix_unit)

    # if "Correlation" in df.columns:
        # change this to r, italicized
    df = df.rename(columns={"Correlation": "<em>r</em>"})
    df = df.rename(columns={"Correlation coefficient": "<em>r</em>"})
    return df.style.hide(axis="index")



from IPython.display import Markdown as md_markdown

def md_basic(x):
    x = x.replace(" 5 m ", " 5m ")
    x = x.replace(" 5 m.", " 5m.")

    x = x.replace("  ", " ")
    # use regex to ensure any numbers have commas
    if "**Figure" in x:
        # ensure the sentence ends with .
        if x[-1] != ".":
            x = x + "."
    if "**Table" in x:
        # ensure the sentence ends with .
        if x[-1] != ".":
            x = x + "."

    x = x.replace(" .", ".")
    x = x.replace(" ,", ",")
    x = x.replace(" :", ":")
    x = x.replace(" ;", ";")
    x = x.replace(" %", "%")

    # ensure there are spaces after commas, using regex
    x = re.sub(r",(\w)", r", \1", x)
    x = x.replace("CO2", "CO<sub>2</sub>")

    return md_markdown(x)

def md(x, number = False):
    x = x.replace("(degC)", "(°C)")
    x = x.replace("(degrees C)", "(°C)")

    #model - observation
    x = x.replace("model - observation", "model-observation")
    #98th, handle this kind of thing appropriately with superscripts
    x = x.replace("98th", "98<sup>th</sup>")
    #2nd
    x = x.replace("2nd", "2<sup>nd</sup>")


    x = x.replace(" 5 m ", " 5m ")
    x = x.replace(" 5 m.", " 5m.")
    if x.lower() == "temperature":
        return "temperature"
    # make CO2 subscript
    x = x.replace("CO2", "CO<sub>2</sub>")
    # fix O_2
    x = x.replace("O_2", "O<sub>2</sub>")
    x = x.replace(" ph ", " pH ")
    x = x.replace(" R2 ", " R<sup>2</sup> ")
    x = x.replace(" R2.", " R<sup>2</sup>.")
    # fix g/kg
    x = x.replace("g/kg", "g kg<sup>-1</sup>")

    # get rid of double spaces
    x = x.replace("  ", " ")
    # use regex to ensure any numbers have commas
    if "**Figure" in x:
        # ensure the sentence ends with .
        if x[-1] != ".":
            x = x + "."
    if "**Table" in x:
        # ensure the sentence ends with .
        if x[-1] != ".":
            x = x + "."
    # ensure there are spaces after commas, using regex
    x = re.sub(r",(\w)", r", \1", x)


    x = x.replace(" .", ".")
    x = x.replace(" ,", ",")
    x = x.replace(" :", ":")
    x = x.replace(" ;", ";")
    x = x.replace(" %", "%")
    # /m^3
    x = x.replace("/m3", "m<sup>-3</sup>")
    x = x.replace("/m^3", "m<sup>-3</sup>")
    # handle /m^2
    x = x.replace("/m2", "m<sup>-2</sup>")
    x = x.replace("/m^2", "m<sup>-2</sup>")
    # handl m-3
    x = x.replace(" m-3", " m<sup>-3</sup>")
    # fix /yr
    x = x.replace("/yr", "year<sup>-1</sup>")
    # fix /day
    x = x.replace("/day", "day<sup>-1</sup>")

    if number:
        if "year" not in x.lower():
            if "period" not in x.lower():
                # do not use numbers between brackets ()
                x = re.sub(r"(\d{1,3})(\d{3})", r"\1,\2", x)
    # identify any urls in the text and convert to markdown links
    url_pattern = re.compile(r'(https?://\S+)')
    urls = url_pattern.findall(x)
    for url in urls:
        x = x.replace(url, f"[{url}]({url})")


    return md_markdown(x)







import pandas as pd
import importlib
import re
import warnings
import calendar
import copy
import oceanval
import pickle

import glob
import geopandas as gpd
import nctoolkit as nc
nc.options(parallel=True)
import os
import pickle
import numpy as np
import xarray as xr
from IPython.display import Markdown as md_markdown
from tqdm import tqdm
import hvplot.pandas
from plotnine import *
from IPython.core.interactiveshell import InteractiveShell
from oceanval.tidiers import tidy_info
InteractiveShell.ast_node_interactivity = "all"

try:
    lon_lim = the_lon_lim
    lat_lim = the_lat_lim
except:
    lon_lim = None
    lat_lim = None
if lon_lim is None:
    lon_lim = [-180, 180]
if lat_lim is None:
    lat_lim = [-90, 90]

try:
    concise = concise_value
except:
    concise = False

# in one line
from oceanval.tidiers import fix_unit, df_display, md, md_basic

warnings.filterwarnings('ignore')

%load_ext rpy2.ipython

test_status = the_test_status


i_figure = 1
i_table = 1
stamp = nc.session_info["stamp"]
out = ".trackers/" + stamp
fast_plot = fast_plot_value
if not os.path.exists(".trackers"):
    os.makedirs(".trackers")
# save out as empty file
with open(out, 'w') as f:
    f.write("")

def bin_value(x, bin_res):
    return np.floor((x + bin_res / 2) / bin_res + 0.5) * bin_res - bin_res / 2

try:
    vv_name = variable
except:
    vv_name = "summary"
    variable = "summary"
Variable = variable.title()



# %% [markdown]

# %% tags=["remove-cell"]

ds_annual = ds_model.copy()
ds_annual.rename({ds_annual.variables[0]: "model"})
ds_annual.append(ds_obs)
ds_annual.tmean()
ds_annual.merge("variables")
ds_annual.rename({ds_obs.variables[0]: "observation"})
out_dir = "../../oceanval_results/annual_mean/"
if not os.path.exists(out_dir):
    os.makedirs(out_dir)
out_file = out_dir + f"annualmean_{variable}_{source}.nc"
ds_annual.to_nc(out_file, zip = True, overwrite = True)
# Calculate the monthly mean and output it
out_def = out_dir + f"annualmean_{variable}_{source}.pkl"
# store definitions
with open(out_def, "wb") as f:
    pickle.dump(definitions, f)

ds_monthly = ds_model.copy()
ds_monthly.rename({ds_monthly.variables[0]: "model"})
ds_monthly.append(ds_obs)
ds_monthly.tmean("month")
ds_monthly.merge("variables", ["month"])
ds_monthly.rename({ds_obs.variables[0]: "observation"})
out_dir = "../../oceanval_results/monthly_mean/"
if not os.path.exists(out_dir):
    os.makedirs(out_dir)
out_file = out_dir + f"monthlymean_{variable}_{source}.nc"
ds_monthly.to_nc(out_file, zip = True, overwrite = True)
# definitions store
out_def = out_dir + f"monthlymean_{variable}_{source}.pkl"
with open(out_def, "wb") as f:
    pickle.dump(definitions, f)

 



from setuptools import Command, find_packages, setup
import sys

DESCRIPTION = "A Python package for ocean model validation and analysis"
LONG_DESCRIPTION = """

**oceanVal** is a Python package designed to automate the process of validating ocean models against observational datasets. It provides a suite of tools to facilitate the comparison of model outputs with various observational data sources, enabling researchers to assess model performance effectively. 

Core abilities of oceanVal include:

  - Matching model output variables to observational datasets 
  - Assesing spatial and temporal performance of ocean models 
  - Assessing model skill using a variety of statistical metrics 
  - Asssessing extent of model biases 
  - Generating comprehensive validation reports in html format


"""


PROJECT_URLS = {
    "Bug Tracker": "https://github.com/pmlmodelling/oceanVal/issues",
    "Source Code": "https://github.com/pmlmodelling/oceanVal",
}

extras_require: dict() = dict()


extras_require["complete"] = ["geoviews", "rioxarray", "cfchecker", "geocube", "geopandas"]

REQUIREMENTS = [i.strip() for i in open("requirements.txt").readlines()]

setup(name='oceanval',
      version='0.1.3',
      description=DESCRIPTION,
      long_description=LONG_DESCRIPTION,
      python_requires='>=3.6.1',
      classifiers=[
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: POSIX :: Linux",
        "Operating System :: MacOS",
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ],

      project_urls=PROJECT_URLS,
      url = "https://github.com/pmlmodelling/oceanVal",
      author='Robert Wilson',
      maintainer='Robert Wilson',
      author_email='rwi@pml.ac.uk',
      include_package_data=True,
      package_data={
      'oceanval': ['data/*'] },

      packages = ["oceanval"],
      setup_requires=[
        'setuptools',
        'setuptools-git',
        'wheel',
    ],
      install_requires = REQUIREMENTS,
      extras_require = extras_require,
      zip_safe=False)







import warnings
from oceanval.parsers import Validator
from oceanval.session import session_info


# function to convert list to string with , separator, with an "and" at the end
def list_to_string(lst):
    if len(lst) == 1:
        return lst[0]
    elif len(lst) == 2:
        return " and ".join(lst)
    else:
        return ", ".join(lst[:-1]) + ", and " + lst[-1]


def ignore_warning(x):
    """
    Parameters
    -------------
    x: str
        Warning message

        Returns
        -------------
        True if the warning should be ignored
        False if the warning should not be ignored
    """
    if "Adding a time series with the same number of time steps" in x:
        return True
    # did not have valid years
    if "did not have valid years" in x:
        return True
    if "There is only file in the dataset. No need to merge" in x:
        return True
    if "time bounds unsupporte" in x:
        return True
    if "deflate" in x:
        return True
    if "None of the points are contained" in x:
        return True
    if "inconsistent variable" in x:
        return True
    if "inconsistent data" in x:
        return True
    if "0 as the fill value" in x:
        return True
    if "found more than one time variabl" in x:
        return True
    if "coordinates variable time" in x and "be assigned" in x:
        return True
    return False



def tidy_warnings(w):
    # A function to tidy up the warnings
    out_warnings = []
    for x in w:
        x_message = str(x.message)
        bad = True
        bad = ignore_warning(x_message) is False
        if bad:
            out_warnings.append(x_message)
    for ww in out_warnings:
        if "months were missing" not in ww:
            warnings.warn(ww)
